@@ -10,10 +10,12 @@
     '(?:release/\\d.\\d[\\x\\d\\.]*)'];
 
   var all_versions = {
-    '3.8': 'dev (3.8)',
+    '3.11': 'dev (3.11)',
+    '3.10': '3.10',
+    '3.9': '3.9',
+    '3.8': '3.8',
     '3.7': '3.7',
     '3.6': '3.6',
-    '3.5': '3.5',
     '2.7': '2.7',
   };
 
@@ -22,6 +24,8 @@
       'fr': 'French',
       'ja': 'Japanese',
       'ko': 'Korean',
+      'pt-br': 'Brazilian Portuguese',
+      'zh-cn': 'Simplified Chinese',
   };
 
   function build_version_select(current_version, current_release) {
@@ -49,6 +53,12 @@
       else
         buf.push('<option value="' + language + '">' + title + '</option>');
     });
+    if (!(current_language in all_languages)) {
+        // In case we're browsing a language that is not yet in all_languages.
+        buf.push('<option value="' + current_language + '" selected="selected">' +
+                 current_language + '</option>');
+        all_languages[current_language] = current_language;
+    }
     buf.push('</select>');
     return buf.join('');
   }
